@@ -1,4 +1,8 @@
-{-# LANGUAGE CPP, FlexibleContexts, Rank2Types #-}
+{-# LANGUAGE BangPatterns, CPP, FlexibleContexts, Rank2Types #-}
+#if __GLASGOW_HASKELL__ >= 704
+{-# OPTIONS_GHC -fsimpl-tick-factor=200 #-}
+#endif
+
 -- |
 -- Module    : Statistics.Function
 -- Copyright : (c) 2009, 2010, 2011 Bryan O'Sullivan
@@ -16,6 +20,7 @@ module Statistics.Function
       minMax
     -- * Sorting
     , sort
+    , gsort
     , sortBy
     , partialSort
     -- * Indexing
@@ -25,19 +30,34 @@ module Statistics.Function
     , nextHighestPowerOfTwo
     -- * Comparison
     , within
+    -- * Arithmetic
+    , square
+    -- * Vectors
+    , unsafeModify
+    -- * Combinators
+    , for
+    , rfor
     ) where
 
 #include "MachDeps.h"
 
+import Control.Monad.ST (ST)
 import Data.Bits ((.|.), shiftR)
 import qualified Data.Vector.Algorithms.Intro as I
 import qualified Data.Vector.Generic as G
+import qualified Data.Vector.Unboxed as U
+import qualified Data.Vector.Unboxed.Mutable as M
 import Statistics.Function.Comparison (within)
 
 -- | Sort a vector.
-sort :: (Ord e, G.Vector v e) => v e -> v e
+sort :: U.Vector Double -> U.Vector Double
 sort = G.modify I.sort
-{-# INLINE sort #-}
+{-# NOINLINE sort #-}
+
+-- | Sort a vector.
+gsort :: (Ord e, G.Vector v e) => v e -> v e
+gsort = G.modify I.sort
+{-# INLINE gsort #-}
 
 -- | Sort a vector using a custom ordering.
 sortBy :: (G.Vector v e) => I.Comparison e -> v e -> v e
@@ -51,7 +71,7 @@ partialSort :: (G.Vector v e, Ord e) =>
             -> v e
             -> v e
 partialSort k = G.modify (`I.partialSort` k)
-{-# INLINE partialSort #-}
+{-# SPECIALIZE partialSort :: Int -> U.Vector Double -> U.Vector Double #-}
 
 -- | Return the indices of a vector.
 indices :: (G.Vector v a, G.Vector v Int) => v a -> v Int
@@ -77,14 +97,52 @@ minMax = fini . G.foldl' go (MM (1/0) (-1/0))
 -- non-negative integer.  If the given value is already a power of
 -- two, it is returned unchanged.  If negative, zero is returned.
 nextHighestPowerOfTwo :: Int -> Int
-nextHighestPowerOfTwo n = o + 1
-    where m = n - 1
-          o = m
-              .|. (m `shiftR` 1)
-              .|. (m `shiftR` 2)
-              .|. (m `shiftR` 4)
-              .|. (m `shiftR` 8)
-              .|. (m `shiftR` 16)
-#if WORD_SIZE_IN_BITS == 64              
-              .|. (m `shiftR` 32)
-#endif                
+nextHighestPowerOfTwo n
+#if WORD_SIZE_IN_BITS == 64
+  = 1 + _i32
+#else
+  = 1 + i16
+#endif
+  where
+    i0   = n - 1
+    i1   = i0  .|. i0  `shiftR` 1
+    i2   = i1  .|. i1  `shiftR` 2
+    i4   = i2  .|. i2  `shiftR` 4
+    i8   = i4  .|. i4  `shiftR` 8
+    i16  = i8  .|. i8  `shiftR` 16
+    _i32 = i16 .|. i16 `shiftR` 32
+-- It could be implemented as
+--
+-- > nextHighestPowerOfTwo n = 1 + foldl' go (n-1) [1, 2, 4, 8, 16, 32]
+--     where go m i = m .|. m `shiftR` i
+--
+-- But GHC do not inline foldl (probably because it's recursive) and
+-- as result function walks list of boxed ints. Hand rolled version
+-- uses unboxed arithmetic.
+
+-- | Multiply a number by itself.
+square :: Double -> Double
+square x = x * x
+
+-- | Simple for loop.  Counts from /start/ to /end/-1.
+for :: Monad m => Int -> Int -> (Int -> m ()) -> m ()
+for n0 !n f = loop n0
+  where
+    loop i | i == n    = return ()
+           | otherwise = f i >> loop (i+1)
+{-# INLINE for #-}
+
+-- | Simple reverse-for loop.  Counts from /start/-1 to /end/ (which
+-- must be less than /start/).
+rfor :: Monad m => Int -> Int -> (Int -> m ()) -> m ()
+rfor n0 !n f = loop n0
+  where
+    loop i | i == n    = return ()
+           | otherwise = let i' = i-1 in f i' >> loop i'
+{-# INLINE rfor #-}
+
+unsafeModify :: M.MVector s Double -> Int -> (Double -> Double) -> ST s ()
+unsafeModify v i f = do
+  k <- M.unsafeRead v i
+  M.unsafeWrite v i (f k)
+{-# INLINE unsafeModify #-}
